'''
    File: test_couchdbsession_unit.py
    Author: Federico Gonzalez
    Description: Tests - CouchDBSession
'''
from mock import Mock, patch
from src.couchdblogger import CouchDBSession, requests
import unittest


class CouchDBSessionTest(unittest.TestCase):

    def setUp(self):
        self.couchdb_session = CouchDBSession()

    def test_is_instance(self):
        self.assertIsInstance(self.couchdb_session, CouchDBSession, "Is instance of CouchDBSession")
        self.assertTrue(issubclass(CouchDBSession, requests.Session), "Is subclass CouchDBSession of requests.Session")

    def test_exception(self):
        self.assertTrue(issubclass(CouchDBSession.CouchDBException, Exception), "Is subclass CouchDBException of Exception")

    def test_request_resp_200_ok(self):
        request_resp = Mock()
        request_resp.status_code = 200
        with patch('requests.Session.request', Mock(return_value=request_resp)) as requests_mock:
            resp = self.couchdb_session.post('http://localhost:5984/_session', data={
                'name': 'username',
                'password': 'password'
            })

            self.assertTrue(requests_mock.called, "")
            self.assertEqual(requests_mock.call_args[0][0], 'POST', "")
            self.assertEqual(requests_mock.call_args[0][1], 'http://localhost:5984/_session', "")
            self.assertIsNotNone(requests_mock.call_args[1]['data'], "")
            self.assertEqual(requests_mock.call_args[1]['data']['name'], 'username', "")
            self.assertEqual(requests_mock.call_args[1]['data']['password'], 'password', "")
            self.assertEqual(len(requests_mock.call_args_list), 1, "")
            self.assertEqual(resp, request_resp, "")

    def test_request_resp_200_ok_with_verify(self):
        request_resp = Mock()
        request_resp.status_code = 200
        self.couchdb_session = CouchDBSession(request_args = {"verify": True})
        with patch('requests.Session.request', Mock(return_value=request_resp)) as requests_mock:
            resp = self.couchdb_session.post('https://localhost:5984/_session', data={
                'name': 'username',
                'password': 'password'
            })

            self.assertTrue(requests_mock.called, "")
            self.assertEqual(requests_mock.call_args[0][0], 'POST', "")
            self.assertEqual(requests_mock.call_args[0][1], 'https://localhost:5984/_session', "")
            self.assertIsNotNone(requests_mock.call_args[1]['data'], "")
            self.assertEqual(requests_mock.call_args[1]['data']['name'], 'username', "")
            self.assertEqual(requests_mock.call_args[1]['data']['password'], 'password', "")
            self.assertTrue(requests_mock.call_args[1]['verify'], "")
            self.assertEqual(len(requests_mock.call_args_list), 1, "")
            self.assertEqual(resp, request_resp, "")

    def test_request_resp_404_not_found(self):
        request_resp = Mock()
        request_resp.status_code = 404
        with patch('requests.Session.request', Mock(return_value=request_resp)) as requests_mock:
            with self.assertRaises(CouchDBSession.CouchDBException):
                self.couchdb_session.post('http://localhost:5984/_session', data={
                'name': 'username',
                'password': 'password'
            })

            self.assertTrue(requests_mock.called, "")
            self.assertEqual(requests_mock.call_args[0][0], 'POST', "")
            self.assertEqual(requests_mock.call_args[0][1], 'http://localhost:5984/_session', "")
            self.assertIsNotNone(requests_mock.call_args[1]['data'], "")
            self.assertEqual(requests_mock.call_args[1]['data']['name'], 'username', "")
            self.assertEqual(requests_mock.call_args[1]['data']['password'], 'password', "")
            self.assertEqual(len(requests_mock.call_args_list), 1, "")

    def test_request_resp_401_unauthorized(self):
        request_resp = Mock()
        request_resp.status_code = 401
        with patch('requests.Session.request', Mock(return_value=request_resp)) as requests_mock:
            with self.assertRaises(CouchDBSession.CouchDBException):
                self.couchdb_session.post('http://localhost:5984/_session', data={
                'name': 'username',
                'password': 'password'
            })

            self.assertTrue(requests_mock.called, "")
            self.assertEqual(requests_mock.call_args[0][0], 'POST', "")
            self.assertEqual(requests_mock.call_args[0][1], 'http://localhost:5984/_session', "")
            self.assertIsNotNone(requests_mock.call_args[1]['data'], "")
            self.assertEqual(requests_mock.call_args[1]['data']['name'], 'username', "")
            self.assertEqual(requests_mock.call_args[1]['data']['password'], 'password', "")
            self.assertEqual(len(requests_mock.call_args_list), 1, "")

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(unittest.TestLoader().loadTestsFromTestCase(CouchDBSessionTest))



'''
    File: test_couchdblogger_unit.py
    Author: Federico Gonzalez
    Description: Tests - CouchDBLogHandler
'''
from mock import Mock, patch
from src.couchdblogger import CouchDBLogHandler, CouchDBSession, logging
import unittest
import json


class CouchDBLogHandlerTest(unittest.TestCase):

    def setUp(self):
        self.couchdb_handler = CouchDBLogHandler()
        self.record = Mock()
        self.record.name = 'process_name'
        self.record.msg = 'log to couchdb'
        self.record.levelname = 'level INFO'
        self.record.created = 1396988156

    def test_is_instance(self):
        self.assertIsInstance(self.couchdb_handler, CouchDBLogHandler, "Is instance of CouchDBLogHandler")
        self.assertTrue(issubclass(CouchDBLogHandler, logging.StreamHandler), "Is subclass CouchDBLogHandler of logging.StreamHandler")

    @patch.object(logging.StreamHandler, '__init__')
    def test_init_username_none(self, *args):
        self.assertFalse(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 5984, "")
        self.assertEqual(self.couchdb_handler.url, 'http://localhost:5984', "")
        self.assertEqual(self.couchdb_handler.database, 'logs', "")
        self.assertEqual(self.couchdb_handler.db_url, 'http://localhost:5984/logs', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")

    @patch.object(logging.StreamHandler, '__init__')
    def test_init_username_none_ssl_true(self, *args):
        self.couchdb_handler = CouchDBLogHandler(ssl=True)
        self.assertTrue(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 5984, "")
        self.assertEqual(self.couchdb_handler.url, 'https://localhost:5984', "")
        self.assertEqual(self.couchdb_handler.database, 'logs', "")
        self.assertEqual(self.couchdb_handler.db_url, 'https://localhost:5984/logs', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")

    @patch.object(logging.StreamHandler, '__init__')
    @patch.object(CouchDBSession, 'get')
    def test_init_username_none_create_database_exist(self, *args):
        self.couchdb_handler = CouchDBLogHandler(create_database=True)
        self.assertTrue(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 5984, "")
        self.assertEqual(self.couchdb_handler.url, 'http://localhost:5984', "")
        self.assertEqual(self.couchdb_handler.database, 'logs', "")
        self.assertEqual(self.couchdb_handler.db_url, 'http://localhost:5984/logs', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")
        self.assertTrue(CouchDBSession.get.called, "")
        self.assertEqual(CouchDBSession.get.call_args[0][0], 'http://localhost:5984/logs', "")

    @patch.object(logging.StreamHandler, '__init__')
    @patch.object(CouchDBSession, 'put')
    def test_init_username_none_create_database_not_exist(self, *args):

        def get_raise(_, url):
            self.assertEqual(url, 'http://localhost:5984/logs', "")
            raise CouchDBSession.CouchDBException
        CouchDBSession.get = get_raise

        self.couchdb_handler = CouchDBLogHandler(create_database=True)
        self.assertTrue(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 5984, "")
        self.assertEqual(self.couchdb_handler.url, 'http://localhost:5984', "")
        self.assertEqual(self.couchdb_handler.database, 'logs', "")
        self.assertEqual(self.couchdb_handler.db_url, 'http://localhost:5984/logs', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")
        self.assertTrue(CouchDBSession.put.called, "")
        self.assertEqual(CouchDBSession.put.call_args[0][0], 'http://localhost:5984/logs', "")

    @patch.object(CouchDBSession, 'post')
    @patch.object(logging.StreamHandler, '__init__')
    def test_init_username_not_is_none(self, *args):
        self.assertFalse(logging.StreamHandler.__init__.called, "")
        self.assertFalse(CouchDBSession.post.called, "")

        self.couchdb_handler = CouchDBLogHandler(host='127.0.0.1', port=8080, database='logs-process', username='user', password='secret')

        self.assertTrue(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 8080, "")
        self.assertEqual(self.couchdb_handler.url, 'http://127.0.0.1:8080', "")
        self.assertEqual(self.couchdb_handler.database, 'logs-process', "")
        self.assertEqual(self.couchdb_handler.db_url, 'http://127.0.0.1:8080/logs-process', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")
        self.assertTrue(CouchDBSession.post.called, "")
        self.assertEqual(CouchDBSession.post.call_args[0][0], 'http://127.0.0.1:8080/_session', "")
        self.assertIsNotNone(CouchDBSession.post.call_args[1]['data'], "")
        self.assertEqual(CouchDBSession.post.call_args[1]['data']['name'], 'user', "")
        self.assertEqual(CouchDBSession.post.call_args[1]['data']['password'], 'secret', "")

    @patch.object(CouchDBSession, 'post')
    @patch.object(logging.StreamHandler, '__init__')
    def test_init_username_not_is_none_and_ssl_true(self, *args):
        self.assertFalse(logging.StreamHandler.__init__.called, "")
        self.assertFalse(CouchDBSession.post.called, "")

        self.couchdb_handler = CouchDBLogHandler(host='127.0.0.1', port=8080, database='logs-process', username='user', password='secret', ssl=True)

        self.assertTrue(logging.StreamHandler.__init__.called, "")
        self.assertEqual(self.couchdb_handler.port, 8080, "")
        self.assertEqual(self.couchdb_handler.url, 'https://user:secret@127.0.0.1:8080', "")
        self.assertEqual(self.couchdb_handler.database, 'logs-process', "")
        self.assertEqual(self.couchdb_handler.db_url, 'https://user:secret@127.0.0.1:8080/logs-process', "")
        self.assertIsInstance(self.couchdb_handler.session, CouchDBSession, "")
        self.assertTrue(CouchDBSession.post.called, "")
        self.assertEqual(CouchDBSession.post.call_args[0][0], 'https://user:secret@127.0.0.1:8080/_session', "")
        self.assertIsNotNone(CouchDBSession.post.call_args[1]['data'], "")
        self.assertEqual(CouchDBSession.post.call_args[1]['data']['name'], 'user', "")
        self.assertEqual(CouchDBSession.post.call_args[1]['data']['password'], 'secret', "")

    def test_format(self):
        left_data = json.loads(self.couchdb_handler.format(self.record))
        right_data = json.loads('{"logger": "process_name", "created": 1396988156, "message": "log to couchdb", "level": "level INFO"}')
        self.assertEqual(left_data, right_data, "")

    @patch.object(CouchDBSession, 'post')
    def test_emit(self, *args):
        self.couchdb_handler.emit(self.record)

        self.assertTrue(CouchDBSession.post.called, "")
        self.assertEqual(CouchDBSession.post.call_args[0][0], 'http://localhost:5984/logs', "")
        self.assertIsNotNone(CouchDBSession.post.call_args[1]['data'], "")
        self.assertIsNotNone(CouchDBSession.post.call_args[1]['headers'], "")

        left_data = json.loads(CouchDBSession.post.call_args[1]['data'])
        right_data = json.loads('{"logger": "process_name", "created": 1396988156, "message": "log to couchdb", "level": "level INFO"}')

        self.assertEqual(left_data, right_data, "")

        self.assertEqual(CouchDBSession.post.call_args[1]['headers']['Content-type'], 'application/json', "")

    def test_new_format(self):

        def format_function(record):
            json_to_post = json.dumps(dict(
                message=record.msg,
                log_level=record.levelname,
                name_logger=record.name,
                extra_message='message',
                log_date_time=record.asctime
            ))
            return json_to_post

        self.couchdb_handler.new_format(format_function)

        self.assertEqual(id(format_function), id(self.couchdb_handler.format), "")

    def test_new_format_with_lambda(self):
        id_format = id(self.couchdb_handler.format)
        self.couchdb_handler.new_format(lambda record: json.dumps(dict(message=record.msg, log_level=record.levelname, name_logger=record.name, extra_message='message', log_date_time=record.asctime)))

        self.assertNotEqual(id_format, id(self.couchdb_handler.format), "")


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(unittest.TestLoader().loadTestsFromTestCase(CouchDBLogHandlerTest))



'''
File: couchdblogger.py
Author: Rinat F Sabitov, Federico Gonzalez
Description:
'''
import logging
import json
import requests


class CouchDBSession(requests.Session):
    """
        CouchDBSession that inherits from requests.Session

        CouchDBSession:
            Session handling orders couchdb

        requests.Session:
            Provides cookie persistence, connection-pooling, and configuration.
    """

    def __init__(self, request_args=None):
        """
            Initialize the couchdb session

        :param request_args; args for request
        """
        super(CouchDBSession, self).__init__()
        self.request_args = request_args or {}

    class CouchDBException(Exception):
        """
            CouchDBException that inherits from Exception

            CouchDBException:
                Exception error when response code greater than 400

            Exception:
                Common base class for all non-exit exceptions.
        """
        pass

    def request(self, *args, **kwargs):
        """
        Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.
        In the case of having a code greater than 400 returns CouchDBException

        Args:
            :param method: method for the new :class:`Request` object.
            :param url: URL for the new :class:`Request` object.

        Kwargs:
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary or bytes to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of 'filename': file-like-objects
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) Float describing the timeout of the
            request.
        :param allow_redirects: (optional) Boolean. Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol to the URL of
            the proxy.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) if ``True``, the SSL cert will be verified.
            A CA_BUNDLE path can also be provided.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair.
        """
        kwargs.update(self.request_args)
        resp = super(CouchDBSession, self).request(*args, **kwargs)
        if resp.status_code >= 400:
            raise self.CouchDBException(resp.text)
        return resp


class CouchDBLogHandler(logging.StreamHandler):
    """
        CouchDBLogHandler that inherits from logging.StreamHandler

        logging.StreamHandler:
            A handler class which writes logging records,
            appropriately formatted, to a stream.
            Note that this class does not close the stream, as
            sys.stdout or sys.stderr may be used.
    """

    def __init__(self, host='localhost', port=5984, database='logs',
        create_database=False, username=None, password=None, ssl=False, request_args=None):
        """
            Initialize the couchdb handler

        :param host: host of couchdb for logging
        :param port: port of couchdb for logging
        :param database: database's name for logging
        :param username: user's name for logging in the database
        :param password: password for logging in the database
        :param create_database: bool to create the database if it does not exist
        :param ssl: bool to use ssl (https)
        :param request_args; json args for request
        """
        super(CouchDBLogHandler, self).__init__()

        self.database = database
        self.port = port
        self.ssl = ssl

        if self.ssl:
            protocol = 'https'
            if username:
                host = "%(username)s:%(password)s@%(host)s" % dict(
                    username=username,
                    password=password,
                    host=host
                )
        else:
            protocol = 'http'

        self.url = "%(protocol)s://%(host)s:%(port)d" % dict(
            protocol=protocol,
            host=host,
            port=port,
        )

        self.db_url = "%(url)s/%(database)s" % dict(
            url=self.url,
            database=database
        )

        self.session = CouchDBSession(request_args=request_args)
        if username:
            self.session.post(self.url+'/_session', data={
                'name': username,
                'password': password
            })

        if create_database:
            try:
                self.session.get(self.db_url)
            except CouchDBSession.CouchDBException:
                self.session.put(self.db_url)

    def new_format(self, format_function):
        """
            Change the format logging

        :param format_function: function to generate couchdb logs from record

        REQUIRED: The function must return a json.dumps to post in couchdb
        """
        self.format = format_function

    def format(self, record):
        """
            Format a logging record to couchdb record

        :param record: loggging record (LogRecord)
        The useful attributes in a LogRecord are described by:

        %(name)s            Name of the logger (logging channel)
        %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
                            WARNING, ERROR, CRITICAL)
        %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
                            "WARNING", "ERROR", "CRITICAL")
        %(pathname)s        Full pathname of the source file where the logging
                            call was issued (if available)
        %(filename)s        Filename portion of pathname
        %(module)s          Module (name portion of filename)
        %(lineno)d          Source line number where the logging call was issued
                            (if available)
        %(funcName)s        Function name
        %(created)f         Time when the LogRecord was created (time.time()
                            return value)
        %(asctime)s         Textual time when the LogRecord was created
        %(msecs)d           Millisecond portion of the creation time
        %(relativeCreated)d Time in milliseconds when the LogRecord was created,
                            relative to the time the logging module was loaded
                            (typically at application startup time)
        %(thread)d          Thread ID (if available)
        %(threadName)s      Thread name (if available)
        %(process)d         Process ID (if available)
        %(message)s         The result of record.getMessage(), computed just as
                            the record is emitted
        """
        return json.dumps(dict(
            message=record.msg,
            level=record.levelname,
            created=record.created,
            logger=record.name
        ))

    def emit(self, record):
        """
            Emit a logging record

        :param record: loggging record
        """
        headers = {'Content-type': 'application/json'}
        self.session.post(self.db_url, data=self.format(record),
            headers=headers
        )
